Require Export ne_list.

Section Domains_and_Relations.

Inductive MD : Type :=
  md1 : Set->MD
| md2 : Set->MD->MD.

Fixpoint etT [M:MD] : Set :=
Cases M of
  (md1 _)   => nat
| (md2 _ m) => nat*(etT m)
end.

Definition etL := [M:MD](ne_list (etT M)).

Fixpoint pow' [s:Set;e:nat] : Set :=
Cases e of
   O     => s
| (S e') => s*(pow' s e')
end.

Definition pow : Set->nat->Set :=
[s:Set;e:nat] Cases e of
   O     => unit
| (S e') => (pow' s e')
end.

Fixpoint arity' [a:Set;M:MD] : nat*(etT M)->Set :=
<[M:MD]nat*(etT M)->Set>
Cases M of
 (md1 b)   => [t:nat*nat] let (ea,eb)=t in
              Cases eb of
                O      => (pow' a ea)
              |(S eb') => (pow' a ea)*(pow' b eb')
              end
|(md2 b m) => [t:nat*nat*(etT m)] let (ea,u)=t in let (eb,v)=u in
              Cases eb of
                O      => (arity' a m (ea,v))
              |(S eb') => (pow' a ea)*(arity' b m (eb',v))
              end
end.

Fixpoint arity [M:MD] : (etT M)->Set :=
<[M:MD](etT M)->Set>
Cases M of 
  (md1 a)   => (pow a) (* =eta [e:nat](pow a e) *)
| (md2 a m) => [t:nat*(etT m)] let (e,u)=t in
               Cases e of
                 O     => (arity m u)
               |(S e') => (arity' a m (e',u))
               end
end.

Fixpoint sar [M:MD] : (etT M)->Set :=
<[M:MD](etT M)->Set>
Cases M of 
  (md1 s)   => (pow s)
| (md2 s m) => [z] let (e,z') = z in
	       Cases e of
	          O	=> (sar m z')
	       | (S e') => (pow' s e')*(sar m z')
	       end
end.

Fixpoint Dslct [k:nat;M:MD] : Set :=
Cases k M of
   O     (md1 s)   => s
|  O     (md2 s _) => s
| (S k') (md2 _ m) => (Dslct k' m)
|  _      _        => unit
end.

End Domains_and_Relations.

Section ManySortedLogic.

Variable M : MD.
Variable L : (etL M).

Definition ar_slct := [n:nat](arity M (ne_nth L n)).

Definition relT := (n:nat)(ar_slct n)->Prop.

Inductive prop : Set :=
  f_atom : (n:nat)(ar_slct n)->prop  
| f_not  : prop->prop
| f_and  : prop->prop->prop
| f_or   : prop->prop->prop
| f_impl : prop->prop->prop
| f_ex   : (k:nat)((Dslct k M)->prop)->prop
| f_all  : (k:nat)((Dslct k M)->prop)->prop.

Definition f_iff := [p,q:prop](f_and (f_impl p q)(f_impl q p)).

Variable REL : relT.

Fixpoint E [p:prop] : Prop :=
Cases p of
  (f_atom n t) => (REL n t)
| (f_not p)    => ~(E p)
| (f_and p q)  => (E p)/\(E q)
| (f_or p q)   => (E p)\/(E q)
| (f_impl p q) => (E p)->(E q)
| (f_ex n dp)  => (EX x:(Dslct n M)|(E (dp x)))
| (f_all n dp) => (x:(Dslct n M))(E (dp x))
end.

End ManySortedLogic.

Syntactic Definition f_Atom := (f_atom ? ?).
Syntactic Definition f_Not  := (f_not ? ?).
Syntactic Definition f_And  := (f_and ? ?).
Syntactic Definition f_Or   := (f_or ? ?).
Syntactic Definition f_Impl := (f_impl ? ?).
Syntactic Definition f_Ex   := (f_ex ? ?).
Syntactic Definition f_All  := (f_all ? ?).
Syntactic Definition f_Iff  := (f_iff ? ?).
Syntactic Definition iE     := (E ? ?).

Definition id : (X:Type)X->X := [X:Type;x:X]x.

Transparent id.

Definition piffp : (A:Prop)A<->A := [A:Prop]<(A->A),(A->A)>{([H:A]H),([H:A]H)}.

