Require Export MSL.

Section Insertion.

Local ar' := arity'.
Local ar := arity.

Definition prod_cons : (s:Set)s->(e:nat)(pow s e)->(pow' s e) :=
[s:Set;x:s;e:nat]
<[e:nat](pow s e)->(pow' s e)>
Cases e of
   O     => [_:unit]x
| (S e') => [z:(pow' s e')](x,z)
end.

Fixpoint prod_snoc' [s:Set;x:s;e:nat] : (pow' s e)->s*(pow' s e) :=
<[e:nat](pow' s e)->s*(pow' s e)>
Cases e of
   O     => [y:s](y,x)
| (S e') => [z:s*(pow' s e')]
	    let (y,z') = z in
	    (y,(prod_snoc' s x e' z'))
end.

Definition prod_snoc : (s:Set)s->(e:nat)(pow s e)->(pow' s e) :=
[s:Set;x:s;e:nat]
<[e:nat](pow s e)->(pow' s e)>
Cases e of
   O     => [_:unit]x
| (S e') => (prod_snoc' s x e')
end.

Fixpoint prod_app' [s:Set;n:nat] : (pow' s n)->(m:nat)(pow s m)
					->(pow' s (plus n m)) :=
<[n:nat](pow' s n)->(m:nat)(pow s m)->(pow' s (plus n m))>
Cases n of
   O     => (prod_cons s)
| (S n') => [u:s*(pow' s n');m:nat;v:(pow s m)]
	    let (x,u')=u in
	    (x,(prod_app' s n' u' m v))
end.

Definition prod_app : (s:Set)(n:nat)(pow s n)->(m:nat)(pow s m)
						->(pow s (plus n m)) :=
[s:Set;n:nat]
<[n:nat](pow s n)->(m:nat)(pow s m)->(pow s (plus n m))>
Cases n of
   O     => [_:unit;m:nat](id (pow s m))
| (S n') => (prod_app' s n')
end.

Fixpoint S_nth_exp [M:MD;n:nat] : (etT M)->(etT M) :=
<[M:MD](etT M)->(etT M)>
Cases M of
  (md1 _)   => [e:nat] Cases n of O => (S e) | _ => e end
| (md2 _ m) => [z:nat*(etT m)] let (e,z') = z in
		Cases n of
		   O     => ((S e),z')
		| (S n') => (e,(S_nth_exp m n' z'))
		end
end.

Fixpoint pred_nth_exp [M:MD;n:nat] : (etT M)->(etT M) :=
<[M:MD](etT M)->(etT M)>
Cases M of
  (md1 _)   => [e:nat] Cases n of O => (pred e) | _ => e end
| (md2 _ m) => [z:nat*(etT m)] let (e,z') = z in
		Cases n of
		   O     => ((pred e),z')
		| (S n') => (e,(pred_nth_exp m n' z'))
		end
end.

Fixpoint etplus [M:MD] : (etT M)->(etT M)->(etT M) :=
<[M:MD](etT M)->(etT M)->(etT M)>
Cases M of
  (md1 _)   => plus
| (md2 _ m) => [u,v:nat*(etT m)]
	       Cases u v of
	       (eu,u') (ev,v') => ((plus eu ev),(etplus m u' v'))
	       end
end.

Fixpoint insertb [M:MD;k:nat] : (Dslct k M)->(z:(etT M))(sar M z)
					->(sar M (S_nth_exp M k z)) :=
<[M:MD](Dslct k M)->(z:(etT M))(sar M z)->(sar M (S_nth_exp M k z))>
Cases M of 
  ((md1 s) as m1)   =>
	<[n:nat](Dslct n m1)->(e:nat)(pow s e)->(pow s (S_nth_exp m1 n e))>
	Cases k of
	  O => (prod_snoc s)
	| _ => [_unit;e:nat](id (sar m1 e))
	end
| ((md2 s m) as m2) =>
	<[k:nat](Dslct k m2)->(z:nat*(etT m))(sar m2 z)
				->(sar m2 (S_nth_exp m2 k z))>
	Cases k of
	   O    =>  [x:s;z:nat*(etT m)]
		    <[z:nat*(etT m)](sar m2 z)->(sar m2 (S_nth_exp m2 O z))>
		    let (e,z') = z in
		    <[e:nat](sar m2 (e,z'))->(pow' s e)*(sar m z')>
		    Cases e of
		       O     => [muqs:(sar m z')](x,muqs)
		    | (S e') => [uqs:(pow' s e')*(sar m z')]
				let (suqs,muqs) = uqs in
				((prod_snoc' s x e' suqs),muqs)
		    end
	| (S k') => [x:(Dslct k' m);z:nat*(etT m)]
		    <[z:nat*(etT m)](sar m2 z)
			->(sar m2 (S_nth_exp m2 (S k') z))>
		    let (e,z') = z in
		    <[e:nat](sar m2 (e,z'))->(sar m2 (e,(S_nth_exp m k' z')))>
		    Cases e of
		       O     => (insertb m k' x z')
		    | (S e') => [uqs:(pow' s e')*(sar m z')]
				let (suqs,muqs) = uqs in
				(suqs,(insertb m k' x z' muqs))
		    end
	end
end.

Fixpoint insertf [M:MD;k:nat] : (Dslct k M)->(z:(etT M))(sar M z)
					->(sar M (S_nth_exp M k z)) :=
<[M:MD](Dslct k M)->(z:(etT M))(sar M z)->(sar M (S_nth_exp M k z))>
Cases M of 
  ((md1 s) as m1)   =>
	<[k:nat](Dslct k m1)->(e:nat)(sar m1 e)->(sar m1 (S_nth_exp m1 k e))>
	Cases k of
	  O => (prod_cons s)
	| _ => [_unit;e:nat](id (sar m1 e))
	end
| ((md2 s m) as m2) =>
	<[k:nat](Dslct k m2)->(z:nat*(etT m))(sar m2 z)
				->(sar m2 (S_nth_exp m2 k z))>
	Cases k of
	   O    =>  [x:s;z:nat*(etT m)]
		    <[z:nat*(etT m)](sar m2 z)->(sar m2 (S_nth_exp m2 O z))>
		    let (e,z') = z in
		    <[e:nat](sar m2 (e,z'))->(pow' s e)*(sar m z')>
		    Cases e of
		       O     => [muqs:(sar m z')](x,muqs)
		    | (S e') => [uqs:(pow' s e')*(sar m z')]
				let (suqs,muqs) = uqs in
				((x,suqs),muqs)
		    end
	| (S k') => [x:(Dslct k' m);z:nat*(etT m)]
		    <[z:nat*(etT m)](sar m2 z)
			->(sar m2 (S_nth_exp m2 (S k') z))>
		    let (e,z') = z in
		    <[e:nat](sar m2 (e,z'))->(sar m2 (e,(S_nth_exp m k' z')))>
		    Cases e of
		       O     => (insertf m k' x z')
		    | (S e') => [uqs:(pow' s e')*(sar m z')]
				let (suqs,muqs) = uqs in
				(suqs,(insertf m k' x z' muqs))
		    end
	end
end.

Fixpoint insertf_tuple [M:MD] : (u:(etT M))(sar M u)->(v:(etT M))(sar M v)
				->(sar M (etplus M u v)) :=
<[M:MD](u:(etT M))(sar M u)->(v:(etT M))(sar M v)->(sar M (etplus M u v))>
Cases M of
  (md1 s)	   => (prod_app s)
| ((md2 s m) as m2) =>
	[u:nat*(etT m)]
	<[u:nat*(etT m)](sar m2 u)->(v:(etT m2))(sar m2 v)
					->(sar m2 (etplus m2 u v))>
	let (eu,u')=u in
	<[eu:nat](sar m2 (eu,u'))->(v:(etT m2))(sar m2 v)
				->(sar m2 (etplus m2 (eu,u') v))>
	Cases eu of
	   O      => [umuqs:(sar m u');v:nat*(etT m)]
		     <[v:nat*(etT m)](sar m2 v)
					->(sar m2 (etplus m2 (O,u') v))>
		     let (ev,v')=v in
		     <[ev:nat](sar m2 (ev,v'))->
				(sar m2 (ev,(etplus m u' v')))>
		     Cases ev of
		        O      => (insertf_tuple m u' umuqs v')
		     | (S ev') => [vuqs:(pow' s ev')*(sar m v')]
				  let (vsuqs,vmuqs)=vuqs in
				  (vsuqs,(insertf_tuple m u' umuqs v' vmuqs))
		     end
	| (S eu') => [uuqs:(pow' s eu')*(sar m u');v:nat*(etT m)]
		     <[uuqs:(pow' s eu')*(sar m u')]
		      (sar m2 v)->(sar m2 (etplus m2 ((S eu'),u') v))>
		     let (usuqs,umuqs)=uuqs in
		     <[v:nat*(etT m)](sar m2 v)
					->(sar m2 (etplus m2 ((S eu'),u') v))>
		     let (ev,v')=v in
		     <[ev:nat](sar m2 (ev,v'))
			->(pow' s (plus eu' ev))*(sar m (etplus m u' v'))>
		     Cases ev of
		       O      => [vmuqs:(sar m v')]
				 ((prod_app' s eu' usuqs O tt),
				  (*artificial beauty*)
				  (insertf_tuple m u' umuqs v' vmuqs))
		    | (S ev') => [vuqs:(pow' s ev')*(sar m v')]
				 let (vsuqs,vmuqs)=vuqs in
				 ((prod_app' s eu' usuqs (S ev') vsuqs),
				  (insertf_tuple m u' umuqs v' vmuqs))
		    end
	end
end.

End Insertion.
