Require Export axiom_types.

Section RB.

Variable M   : MD.
Variable L   : (etL M).
Variable Rel : (relT M L).
Variable ac  : AC.
Variable ned : (NED M).

Local prp := (prop M L).
Local L2D:=(nel2disj M L).
Local D2L:=(disj2nel M L).
Local L2C:=(nel2conj M L).
Local C2L:=(conj2nel M L).


Remark or2app :
 (A,B:(ne_list prp))(iE Rel (L2D A))\/(iE Rel (L2D B))
  ->(iE Rel (L2D (ne_app A B))).
Proof.
Induction A.
Trivial.
Intros a t IH B H.
(Elim H; Intro H0).
(Elim H0; Intro H1).
Left. Exact H1.
Right. Apply (IH B).
Left. Exact H1.
Right. Apply (IH B).
Right. Exact H0.
Qed.

Remark app2or :
 (A,B:(ne_list prp))(iE Rel (L2D (ne_app A B)))
  ->(iE Rel (L2D A))\/(iE Rel (L2D B)).
Proof.
Induction A.
Trivial.
Intros a t IH B H.Elim H; Intro H0.
Left. Left. Exact H0.
Elim (IH B H0); Intro H1.
Left. Right. Exact H1.
Right. Exact H1.
Qed.

Lemma and2app :
 (A,B:(ne_list prp))(iE Rel (L2C A))/\(iE Rel (L2C B))
  ->(iE Rel (L2C (ne_app A B))).
Proof.
Induction A.
Trivial.
Intros a A' IH B H.
Elim H. Intros H1 H2.
Elim H1. Intros H3 H4.
Split. Exact H3.
Apply IH. Split. Exact H4. Exact H2.
Qed.

Lemma app2and :
 (A,B:(ne_list prp))(iE Rel (L2C (ne_app A B)))
  ->(iE Rel (L2C A))/\(iE Rel (L2C B)).
Proof.
Induction A.
Trivial.
Intros a A' IH B H. Elim H. Intros H1 H2.
Elim (IH B H2). Intros H3 H4.
Split. Split. Exact H1. Exact H3. Exact H4.
Qed.

Remark rbdisj_remark1 : (p:prp)(iE Rel p)->(iE Rel (rbdisj M L p)).
Proof.
Induction p; Trivial.
Intros p1 IH1 p2 IH2 H.
Apply (or2app (D2L p1)(D2L p2)).
(Elim H; Intro H0).
Left. Exact (IH1 H0).
Right. Exact (IH2 H0).
Qed.

Remark rbdisj_remark2 : (p:prp)(iE Rel (rbdisj M L p))->(iE Rel p).
Proof.
Induction p; Trivial.
Intros p1 IH1 p2 IH2 H.
Elim (app2or (disj2nel M L p1) (disj2nel M L p2) H); Intro H0.
Left. Exact (IH1 H0).
Right. Exact (IH2 H0).
Qed.

Remark nel_rbdisj_remark1 :
 (p:prp)(iE Rel p)->(iE Rel (nel2conj M L (nel_rbdisj M L p))).
Proof.
Induction p; Trivial.
Intros p1 IH1 p2 IH2 H.
Elim H. Intros H1 H2.
Simpl. Apply and2app.
Split.
Apply IH1. Exact H1.
Apply IH2. Exact H2.
Intros p1 IH1 p2 IH2 H.
Simpl. Apply rbdisj_remark1.
Exact H.
Qed.

Remark nel_rbdisj_remark2 :
 (p:prp)(iE Rel (nel2conj M L (nel_rbdisj M L p)))->(iE Rel p).
Proof.
Induction p; Trivial.
Intros p1 IH1 p2 IH2 H.
Elim (app2and (nel_rbdisj M L p1)(nel_rbdisj M L p2) H).
Intros H1 H2. Split. Exact (IH1 H1). Exact (IH2 H2).
Intros p1 IH1 p2 IH2 H.
Apply rbdisj_remark2. Exact H.
Qed.

Lemma L2C2L : (p:prp)(iE Rel p)->(iE Rel (L2C (C2L p))).
Proof.
Induction p; Trivial.
Intros p1 IH1 p2 IH2 H.
Elim H. Intros H1 H2.
Simpl. Apply and2app.
Split. Exact (IH1 H1). Exact (IH2 H2).
Qed.

Theorem rb_thm1 : (p:prp)(iE Rel p)->(iE Rel (rb M L p)).
Proof.
Induction p; Trivial.
Intros p1 IH1 p2 IH2 H. Unfold rb. Apply nel_rbdisj_remark1. Exact H.
Intros p1 IH1 p2 IH2 H. Unfold rb. Apply nel_rbdisj_remark1. Exact H.
Simpl. Intros k p' IH H x. Exact (IH x (H x)).
Qed.

Theorem rb_thm2 : (p:prp)(iE Rel (rb M L p))->(iE Rel p).
Proof.
Induction p; Trivial.
Intros p1 IH1 p2 IH2 H. Apply nel_rbdisj_remark2. Exact H.
Intros p1 IH1 p2 IH2 H. Apply nel_rbdisj_remark2. Exact H.
Simpl. Intros k p' IH H x. Exact (IH x (H x)).
Qed.

End RB.
