Require Export sugarCLAUS.
Require Export bliksem.
Require Export tptp.
Require Export many2one.

Section Newmans_Lemma.

Variable em : EM. (* Excluded Middle *)
Variable ac : AC. (* Axiom of Choice *)
Variable  A : Set.
Variable  a : A.
Variable  R : A->A->Prop.

Local sort0 := A.
Local sort1 := nat.
Local sort2 := nat->A.

(* list of domains *)
Local M := (md2 sort0 (md2 sort1 (md1 sort2))).

(* non-emptiness of all domains *)
Local ned := (mkned M (a,(0,[_]a))).

(* Reflexive and transitive closure of R *)
Inductive RR : A->A->Prop :=
  refl_RR  : (x:A)(RR x x)
| trans_RR : (x,y,z:A)(R x y)->(RR y z)->(RR x z).

(* unCurrying *)
Local R'  := [p:A*A] let (x,y)=p in (R x y).
Local RR' := [p:A*A] let (x,y)=p in (RR x y).
Local eq' := [p:A*A] let (x,y)=p in (eq sort0 x y).

(* list of arities of R', RR' and eq' *)
Local L := (ne_cons (2,(0,0))(ne_cons (2,(0,0))(sngl (2,(0,0))))).

(* pi-product packaging R', RR' and eq' *)
Local Rel : (relT M L) :=
 [i]<[i:nat](ar_slct M L i)->Prop>
 Cases i of
 |(0) => R'
 |(1) => RR'
 | _  => eq'
end.

(* The interpretation-function in (prop M L)->Prop *)
Local interpret := (iE Rel).

(* Curried object-level relations *)
Local f_R  := [x,y:A](f_atom M L (0)(x,y)).
Local f_RR := [x,y:A](f_atom M L (1)(x,y)).
Local f_eq := [x,y:A](f_atom M L (2)(x,y)).

(* OBJECT LEVEL *)

(* Weak Church Rosser *)
Local f_WCR :=
 (!ALL(0) x:sort0 | (!ALL(0) y:sort0 | (!ALL(0) z:sort0 |
  ((f_R x y) !/\ (f_R x z))
   !-> (!EX(0) u:sort0 | ((f_RR y u) !/\ (f_RR z u))) ))).

(* The property of an sort0-element being Church Rosser *)
Local f_CR :=
 [x:sort0](!ALL(0) y:sort0 | (!ALL(0) z:sort0 |
  ((f_RR x y) !/\ (f_RR x z))
   !-> (!EX(0) u:sort0 | ((f_RR y u) !/\ (f_RR z u))) )).

(* The existence of an non-CR element *)
Local f_ne_nCR := (!EX(0) x:sort0 | !~(f_CR x)).
 
(* Dependent Choice for the subset of A of non-CR elements *)
Local f_DC :=
  (!ALL(0) x:sort0 | !~(f_CR x) !-> (!EX(0) y:sort0 | !~(f_CR y) !/\ (f_R x y)))
  !-> f_ne_nCR
   !->(!EX(2) f:sort2 |
       (!ALL(1) n:sort1 | !~(f_CR (apply (type sort2 f)(type sort1 n))))
        !/\ (!ALL(1) n:sort1 | (f_R (apply (type sort2 f)(type sort1 n))
                                  (apply (type sort2 f)(type sort1 (S n)))))).

(* Strong Normalisation *)
Local f_SN :=
 !~(!EX(2) f:sort2 |
   (!ALL(1) n:sort1 | (f_R (apply (type sort2 f)(type sort1 n))
                         (apply (type sort2 f)(type sort1 (S n)))))).

(* If there is a non-CR element, then there is a minimal such element. *)
Local f_MP :=
 f_ne_nCR 
  !->(!EX(0) x:sort0 |!~(f_CR x)
     !/\ !~(!EX(0) y:sort0 | !~(f_CR y) !/\ (f_R x y))).

(* Newman's Lemma *)

Local f_NL := f_SN !-> f_DC !-> f_WCR !-> (!ALL(0) x:sort0 | (f_CR x)).


(* Our proof by hand showed that we need the following properties of RR,
   which follow from its inductive definition above. (See file newman.v.)
   We shall have to give these (in clausal form) to Bliksem. *)

Local f_RRrefl := (!ALL(0) x:sort0 | (f_RR x x)).

Local f_RRtrans :=
 (!ALL(0) x:sort0 | (!ALL(0) y:sort0 | (!ALL(0) z:sort0 |
  (f_RR x y) !-> (f_RR y z) !-> (f_RR x z)))).

Local f_RR01 :=
 (!ALL(0) x:sort0 | (!ALL(0) y:sort0 |
  (f_RR x y)
   !-> (f_eq (type sort0 x)(type sort0 y))
        !\/ (!EX(0) z:sort0 | (f_R x z) !/\ (f_RR z y)))).

Local f_NLplus :=
 f_RRrefl !-> f_RRtrans !-> f_RR01 !-> f_NL.

(*****************************************************************************)


Remark NLplus : (interpret f_NLplus).
Proof.
Apply (dcf Rel em ac ned f_NLplus).
Intro skol.
Time Lazy Beta Delta -[apply type not sort0 sort1 sort2] Iota.

Bliksem "nl.blk".
Bliksem Add "nl.blk" shift "skol".
TPTP "nl.p".
TPTP Add "nl.p" shift "skol".

Abort.
End Newmans_Lemma.


