open Pfedit
open Names
open Trad
open Mach
open Pp
open Printer
open Vernacinterp
open Std
open Term
open Generic
open Impuniv
open Ast
open Termenv
open Pp_control
open Tacmach
open Tactics

let constant sp id =
  Machops.global_reference (gLOB (initial_sign ())) (path_of_string sp)
    (id_of_string id)

let coq_not_lazy = lazy (constant "#Logic#not.cci" "not")
let coq_or_lazy = lazy (constant "#Logic#or.cci" "or")
let coq_O_lazy = lazy (constant "#Datatypes#nat.cci" "O")
let coq_S_lazy = lazy (constant "#Datatypes#nat.cci" "S")
let coq_pair_lazy = lazy (constant "#Datatypes#prod.cci" "pair")
(* let coq_tt_lazy = lazy (constant "#Datatypes#unit.cci" "tt") *)
(* let coq_type_lazy = lazy (constant "#newman#type.cci" "type") *)
let coq_eq_lazy = lazy (constant "#Logic#eq.cci" "eq")

let rec print_bliksem ft t  =

  let coq_not = Lazy.force coq_not_lazy
  and coq_or = Lazy.force coq_or_lazy
  and coq_O = Lazy.force coq_O_lazy
  and coq_S = Lazy.force coq_S_lazy
  and coq_pair = Lazy.force coq_pair_lazy
  (* and coq_tt = Lazy.force coq_tt_lazy *)
  (* and coq_type = Lazy.force coq_type_lazy *)
  and coq_eq = Lazy.force coq_eq_lazy in

  let print_unknown t =
    [< 'sTR "\"" ; prterm t ; 'sTR "\"" >] in

  let print_id_upper i =
    [< 'sTR (String.uppercase (string_of_id i)) >] in

  let rec eval_nat t =
    if eq_constr t coq_O then 0
    else if isAppL t then
      let a = args_app t in
	if eq_constr (hd_app t) coq_S & Array.length a > 0 then
	  1 + eval_nat a.(0)
	else raise (Failure "eval_nat")
    else raise (Failure "eval_nat") in

  let print_nat t =
    [< 'iNT (eval_nat t) >] in
    
  let add_context n l =
    (match n with
	 Name i -> next_global_ident_away i l
       | Anonymous -> id_of_string "_") :: l in

  let filter_implicits t a =
    let rec filter l n b =
      match b with
	  c :: d ->
	    (match l with
		 m :: o ->
		   if m = n then filter o (n + 1) d
		   else c :: (filter l (n + 1) d)
	       | [] -> b)
	| [] -> [] in
      match t with
	  DOPN(Const _, _) -> filter (constant_implicits t) 1 a
	| _ -> a in

  let rec print_constr_list v x = function
      [ ] -> [< >]
    | [ t ] -> [< print_constr v t >]
    | t :: (u :: a) ->
	[< print_constr v t ; x ; print_constr_list v x (u :: a) >]

  and print_constr v t = match t with
      DOPN(AppL, a) ->
	let f = a.(0) in
	  if eq_constr f coq_or & Array.length a = 3 then
	    [< print_constr v a.(1) ;  'fNL ;
	       'sTR "," ; print_constr v a.(2) >]
	  else if eq_constr f coq_not & Array.length a = 2 then
	    [< 'sTR "!" ; print_constr v a.(1) >]
	  else if eq_constr f coq_S then
	    try let n = eval_nat t in print_nat t
	    with Failure "eval_nat" ->
	      [< 'sTR "s(" ; print_constr v a.(1) ; 'sTR ")" >]
	  else if eq_constr f coq_pair & Array.length a = 5 then
	    (* if eq_constr a.(4) coq_tt then
	      print_constr v a.(3)
	    else *)
	      hOV 0 [< print_constr v a.(3) ; 'sTR "," ; 'bRK(0,0) ;
		 print_constr v a.(4) >]
	  (* else if eq_constr f coq_type & Array.length a = 3 then
	    [< 'sTR "type(" ; hOV 0 [< prterm a.(1) ; 'sTR "," ; 'bRK(0,0) ;
	       print_constr v a.(2) >] ; 'sTR ")" >] *)
	  else if eq_constr f coq_eq & Array.length a = 4 then
	    [< hOV 0 [< print_constr v a.(2) ; 'sTR "=" ; 'bRK(0,0) ;
	       print_constr v a.(3) >] ; >]
	  else
	    let l = List.tl (Array.to_list a) in
	      [< print_constr v a.(0) ; 'sTR "(" ;
		 hOV 0 (print_constr_list v [< 'sTR "," ; 'bRK(0,0) >]
			  (filter_implicits a.(0) l)) ;
		 'sTR ")" >]
    | DOPN(_, _) ->
	if eq_constr t coq_O then
	  print_nat t
	else
	  prterm t
    | DOP2(o, _, u) ->
	(match o with
	     Prod -> print_constr v u
	   | _ -> print_unknown t)
    | DLAM(n, u) ->
	print_constr (add_context n v) u
    | VAR i -> print_id i
    | Rel n -> [< print_id_upper (List.nth v (n - 1)) >]
    | _ -> print_unknown t in

    mSG_with ft [< 'sTR " {    " ;
		   hOV 1 [< print_constr [] t >] ; 'fNL ;
		   'sTR " }." ; 'fNL >]
      
let bliksem ft a =
  let (ev, si) = get_evmap_sign None in
  let t = (constr_of_com ev si a) in
    match t with
	DOPN(Const _, [| |]) -> print_bliksem ft (const_value ev t)
      | VAR _ -> print_bliksem ft (type_of ev si t)
      | _ -> print_bliksem ft t

let rec bliksem_hypos ft a =
  let b = strip_outer_cast a in
    match kind_of_term b with
	IsProd(_, d, r) -> print_bliksem ft d; bliksem_hypos ft r;
      | _ -> ()

let bliksem_goal ft =
  let a = pf_concl (nth_goal_of_pftreestate 1 (get_pftreestate ())) in
    bliksem_hypos ft a

let file_fd x n =
  Unix.openfile n
    (if x = 0 then [ Unix.O_WRONLY; Unix.O_TRUNC; Unix.O_CREAT ]
     else [ Unix.O_WRONLY; Unix.O_APPEND; Unix.O_CREAT ])
    0o666

let file_ft fd =
  let ch = Unix.out_channel_of_descr fd in
    Format.formatter_of_out_channel ch

let _ =
  vinterp_add
    ("bliksem",
     function
	 [VARG_COMMAND a] ->
	   (fun () -> bliksem std_ft a)
       | [VARG_NUMBER x; VARG_STRING n; VARG_COMMAND a] ->
	   (fun () -> let fd = file_fd x n in let ft = file_ft fd in
	      bliksem ft a; Unix.close fd)
       | [] ->
	   (fun () -> bliksem_goal std_ft)
       | [VARG_NUMBER x; VARG_STRING n] ->
	   (fun () -> let fd = file_fd x n in let ft = file_ft fd in
	      bliksem_goal ft; Unix.close fd)
       | _ -> anomaly "Bliksem")

