Require Export MSL.
Require Export insert.

Chapter CLAUSIFICATION.

Variable M : MD.
Variable L : (etL M).
Local prp := (prop M L).

(* NEGATION NORMAL FORM *)

Inductive pol : Set :=
 pos : pol | neg : pol.

Fixpoint nnf [p:prp] : pol->prp :=
[a] Cases p a of
 (f_not p0)     pos => (nnf p0 neg)
|(f_not p0)     neg => (nnf p0 pos)
|(f_and p1 p2)  pos => (f_And (nnf p1 pos)(nnf p2 pos))
|(f_and p1 p2)  neg => (f_Or  (nnf p1 neg)(nnf p2 neg))
|(f_or p1 p2)	pos => (f_Or  (nnf p1 pos)(nnf p2 pos))
|(f_or p1 p2)   neg => (f_And (nnf p1 neg)(nnf p2 neg))
|(f_impl p1 p2) pos => (f_Or  (nnf p1 neg)(nnf p2 pos))
|(f_impl p1 p2) neg => (f_And (nnf p1 pos)(nnf p2 neg))
|(f_ex k dp)    pos => (f_Ex  k [x:(Dslct k M)](nnf (dp x) pos))
|(f_ex k dp)    neg => (f_All k [x:(Dslct k M)](nnf (dp x) neg))
|(f_all k dp)   pos => (f_All k [x:(Dslct k M)](nnf (dp x) pos))
|(f_all k dp)   neg => (f_Ex  k [x:(Dslct k M)](nnf (dp x) neg))
| atom          pos => atom
| atom          neg => (f_Not atom)
end.

(* PRENEX NORMAL FORM *)

Fixpoint pnf_conj_r [p,q:prp] : prp :=
Cases q of
 (f_ex k dq)  => (f_Ex  k [x:(Dslct k M)](pnf_conj_r p (dq x)))
|(f_all k dq) => (f_All k [x:(Dslct k M)](pnf_conj_r p (dq x)))
| q'          => (f_And p q')
end.

Fixpoint pnf_conj [p:prp] : prp->prp :=
[q:prp] Cases p of
 (f_ex k dp)  => (f_Ex  k [x:(Dslct k M)](pnf_conj (dp x) q))
|(f_all k dp) => (f_All k [x:(Dslct k M)](pnf_conj (dp x) q))
| p'	      => (pnf_conj_r p' q)
end.

Fixpoint pnf_disj_r [p,q:prp] : prp :=
Cases q of
 (f_ex k dq)  => (f_Ex  k [x:(Dslct k M)](pnf_disj_r p (dq x)))
|(f_all k dq) => (f_All k [x:(Dslct k M)](pnf_disj_r p (dq x)))
| q'          => (f_Or p q')
end.

Fixpoint pnf_disj [p:prp] : prp->prp :=
[q:prp]
Cases p of
 (f_ex k dp)  => (f_Ex  k [x:(Dslct k M)](pnf_disj (dp x) q))
|(f_all k dp) => (f_All k [x:(Dslct k M)](pnf_disj (dp x) q))
| p'	      => (pnf_disj_r p' q)
end.

Fixpoint pnf [p:prp] : prp :=
Cases p of
 (f_and p1 p2) => (pnf_conj (pnf p1)(pnf p2))
|(f_or p1 p2)  => (pnf_disj (pnf p1)(pnf p2))
|(f_ex k dp)   => (f_Ex  k [x:(Dslct k M)](pnf (dp x)))
|(f_all k dp)  => (f_All k [x:(Dslct k M)](pnf (dp x)))
| p'	       => p'
end.

(* CONJUNCTIVE NORMAL FORM *)

Fixpoint distr_r [p,q:prp] : prp :=
Cases q of
 (f_and q1 q2) => (f_And (distr_r p q1)(distr_r p q2))
| q'           => (f_Or p q')
end.

Fixpoint distr [p:prp] : prp->prp :=
[q:prp] Cases p of
|(f_and p1 p2) => (f_And (distr p1 q)(distr p2 q))
| p'           => (distr_r p' q)
end.

Fixpoint cnf [p:prp] : prp :=
Cases p of
 (f_and p1 p2) => (f_And (cnf p1)(cnf p2))
|(f_or p1 p2)  => (distr (cnf p1)(cnf p2))
|(f_ex k dp)   => (f_Ex k [x:(Dslct k M)](cnf (dp x)))
|(f_all k dp)  => (f_All k [x:(Dslct k M)](cnf (dp x)))
| p'           => p'
end.

(* RIGHT-BRANCHING DISJUNCTIONS & CONJUNCTIONS *)

Fixpoint disj2nel [p:prp] : (ne_list prp) :=
Cases p of
 (f_or p1 p2) => (ne_app (disj2nel p1)(disj2nel p2))
| p'          => (sngl p')
end.

Fixpoint nel2disj [L:(ne_list prp)] : prp :=
Cases L of
  (sngl p)	=> p
| (ne_cons p t) => (f_Or p (nel2disj t))
end.

Definition rbdisj : prp->prp := [p](nel2disj (disj2nel p)).

Fixpoint conj2nel [p:prp] : (ne_list prp) :=
Cases p of
 (f_and p1 p2) => (ne_app (conj2nel p1)(conj2nel p2))
| p'           => (sngl p')
end.

Fixpoint nel2conj [L:(ne_list prp)] : prp :=
Cases L of
 (sngl p)      => p
|(ne_cons p t) => (f_And p (nel2conj t))
end.

(* returns non-empty list of right-associative disjunctions *)
Fixpoint nel_rbdisj [p:prp] : (ne_list prp) :=
Cases p of
 (f_and p1 p2) => (ne_app (nel_rbdisj p1)(nel_rbdisj p2))
| p'           => (sngl (rbdisj p'))
end.

Fixpoint rb [p:prp] : prp :=
Cases p of
 (f_all k p') => (f_All k [x:(Dslct k M)](rb (p' x)))
| p'          => (nel2conj (nel_rbdisj p'))
end.

(* SKOLEMIZATION *)

Fixpoint zeroes [M:MD] : (etT M) :=
<[M:MD](etT M)> Cases M of 
  (md1 _)   =>  O
| (md2 _ m) => (O,(zeroes m))
end.

Fixpoint mktt [M:MD] : (sar M (zeroes M)) :=
<[M:MD](sar M (zeroes M))>
Cases M of 
  (md1 _)   => tt
| (md2 _ m) => (mktt m)
end.

Definition skolT : Set := nat->nat->(k:nat)(t:(etT M))(sar M t)->(Dslct k M).

Variable skol : skolT.

Fixpoint sklm_rec [i,j:nat;z:(etT M);t:(sar M z);p:prp] : prp :=
Cases p of
 (f_ex k dp)  => (sklm_rec i (S j) z t (dp (skol i j k z t)))
|(f_all k dp) => (f_All k [x:(Dslct k M)]
		   (sklm_rec i j (S_nth_exp M k z)(insertb M k x z t)(dp x)))
| q           => q
end.

Definition sklm := [i:nat](sklm_rec i O (zeroes M)(mktt M)).

(* DISTRIBUTION OF UNIVERSAL QUANTIFIERS OVER CLAUSES *)

Fixpoint duqc1 [p:prp] : prp :=
Cases p of
 (f_all k dp) => (f_All k [x:(Dslct k M)](duqc1 (dp x)))
|(f_and p' _) => p'
| p'          => p'
end.

Fixpoint duqc2 [p:prp] : prp :=
Cases p of
 (f_all k dp) => (f_All k [x:(Dslct k M)](duqc2 (dp x)))
|(f_and _ p') => p'
| p'          => p'
end.

Fixpoint duqc_rec [p:prp;n:nat] : (ne_list prp) :=
Cases n of
  O     => (sngl p)
|(S n') => (ne_cons (duqc1 p)(duqc_rec (duqc2 p) n'))
end.

Variable ned : (k:nat)(Dslct k M).

Fixpoint cdepth [p:prp] : nat :=
Cases p of
 (f_and _ p') => (S (cdepth p'))
|(f_all k dp) => (cdepth (dp (ned k)))
| _           => O
end.

Definition duqc : prp->(ne_list prp) := 
[p] Cases p of
 ((f_all _ _) as p') => (duqc_rec p' (cdepth p'))
| p'                 => (sngl p')
end.

(* CLAUSIFYING CONJUNCTS SEPARATELY *)

Definition dscp : nat->prp->(ne_list prp) :=
 [i;p](duqc (rb (sklm i (cnf (pnf p))))).

Fixpoint dscp_rec [i:nat;L:(ne_list prp)] : (ne_list prp) :=
Cases L of
 (sngl p)      => (dscp i p)
|(ne_cons p t) => (ne_app (dscp i p)(dscp_rec (S i) t))
end.

(* IMPLICATIVE FORMS *)

Variable Rel : (relT M L).

(* maps directly on Prop as we have no formal counterpart of False *)
Fixpoint cif [op:prp->Prop;L:(ne_list prp)] : Prop :=
Cases L of
 (sngl p)      => (op p)->False
|(ne_cons p t) => (op p)->(cif op t)
end.

Definition chpol := [p:prp]
Cases p of
 (f_not p') => p'
| p'	    => (f_Not p')
end.

Fixpoint disj2impl [p:prp] : Prop := 
Cases p of
 (f_or p1 p2) => (iE Rel (chpol p1))->(disj2impl p2)
|(f_all k p') => (x:(Dslct k M))(disj2impl (p' x))
| p'          => (iE Rel (chpol p'))->False
end.

(* MINIMAL CLAUSAL FORM *)

Definition mcf0 : prp->Prop :=
 [p](cif disj2impl (dscp_rec O (conj2nel (cnf (nnf p neg))))).

(* DISJUNCTIVE CLAUSAL FORM *)

Definition dcf0 : prp->Prop :=
 [p](cif (iE Rel)(dscp_rec O (conj2nel (cnf (nnf p neg))))).


End CLAUSIFICATION.
