Require sugarCLAUS.

(* Section Newmans_Lemma. *)

Variable em : EM. (* Excluded Middle *)
Variable ac : AC. (* Axiom of Choice *)
Variable  A : Set.
Variable  a : A.
Variable  R : A->A->Prop.

(* list of domains *)
Local M := (md2 A (md2 nat (md1 nat->A))).

(* non-emptiness of all domains *)
Local ned := (mkned M (a,(0,[_]a))).

(* Reflexive and transitive closure of R *)
Inductive RR : A->A->Prop :=
  refl_RR  : (x:A)(RR x x)
| trans_RR : (x,y,z:A)(R x y)->(RR y z)->(RR x z).

(* unCurrying *)
Local R'  := [p:A*A] let (x,y)=p in (R x y).
Local RR' := [p:A*A] let (x,y)=p in (RR x y).
Local eq' := [p:A*A] let (x,y)=p in (eq A x y).

(* list of arities of R', RR' and eq' *)
Local L := (ne_cons (2,(0,0))(ne_cons (2,(0,0))(sngl (2,(0,0))))).

(* pi-product packaging R', RR' and eq' *)
Local Rel : (relT M L) :=
 [i]<[i:nat](ar_slct M L i)->Prop>
 Cases i of
 |(0) => R'
 |(1) => RR'
 | _  => eq'
end.

(* The interpretation-function in (prop M L)->Prop *)
Local interpret := (iE Rel).

(* Curried object-level relations *)
Local f_R  := [x,y:A](f_atom M L (0)(x,y)).
Local f_RR := [x,y:A](f_atom M L (1)(x,y)).
Local f_eq := [x,y:A](f_atom M L (2)(x,y)).

(* Because Bliksem is one-sorted and doesn't allow quantification over
   function symbols, we employ the following trick. *)
Local apply  := [f:nat->A;n:nat](f n).

(* Polymorphic identity is used to reconstruct the types more easily. *)
Local type := [X:Set;x:X]x.


(* Weak Church Rosser *)
Local f_WCR :=
 (!ALL(0) x:A | (!ALL(0) y:A | (!ALL(0) z:A |
  ((f_R x y) !/\ (f_R x z))
   !-> (!EX(0) u:A | ((f_RR y u) !/\ (f_RR z u))) ))).

(* The property of an A-element being Church Rosser *)
Local f_CR :=
 [x:A](!ALL(0) y:A | (!ALL(0) z:A |
  ((f_RR x y) !/\ (f_RR x z))
   !-> (!EX(0) u:A | ((f_RR y u) !/\ (f_RR z u))) )).

(* The existence of an non-CR element *)
Local ne_nCR := (!EX(0) x:A | !~(f_CR x)).
 
(* Dependent Choice for the subset of A of non-CR elements *)
Local f_DC :=
 ne_nCR !->
 (!ALL(0) x:A | !~(f_CR x) !-> (!EX(0) y:A | !~(f_CR y) !/\ (f_R x y)))
  !-> (!EX(2) f:nat->A |
       (!ALL(1) n:nat | !~(f_CR (apply (type nat->A f)(type nat n))))
        !/\ (!ALL(1) n:nat | (f_R (apply (type nat->A f)(type nat n))
                                  (apply (type nat->A f)(type nat (S n)))))).

(* Strong Normalisation *)
Local f_SN :=
 !~(!EX(2) f:nat->A |
   (!ALL(1) n:nat | (f_R (apply (type nat->A f)(type nat n))
                         (apply (type nat->A f)(type nat (S n)))))).

(* If there is a non-CR element, then there is a minimal such element;
  the statement is a reduct of (interpret ne_nCR !->(!EX(0) x:A |(f_MP x))). *)
Local f_MP :=
 ne_nCR 
  !->(!EX(0) x:A |!~(f_CR x) !/\ !~(!EX(0) y:A | !~(f_CR y) !/\ (f_R x y))).

(* Newman's Lemma *)

Local f_NL := f_SN !-> f_DC !-> f_WCR !-> (!ALL(0) x:A | (f_CR x)).


(* Our proof by hand showed that we need the following properties of RR,
   which follow from its inductive definition above.
   We shall have to give these (in clausal form) to Bliksem. *)

Local f_RRrefl := (!ALL(0) x:A | (f_RR x x)).

Local f_RR01 :=
 (!ALL(0) x:A | (!ALL(0) y:A |
  (f_RR x y) !-> (f_eq x y) !\/ (!EX(0) z:A | (f_R x z) !/\ (f_RR z y)))).

Local f_RRtrans :=
 (!ALL(0) x:A | (!ALL(0) y:A | (!ALL(0) z:A |
  (f_RR x y) !-> (f_RR y z) !-> (f_RR x z)))).


Local f_NLplus :=
 f_RRrefl !-> f_RRtrans !-> f_RR01 !-> f_NL.

(*****************************************************************************)

(*
Lemma NLplus : (interpret f_NLplus).
Proof.
Apply (dcf Rel em ac ned f_NLplus).

(**
Time Cbv Beta Delta -[skolT apply type] Iota.
(* (15.37u,0.14s) *)
**)

Time Lazy Beta Delta -[skolT apply type not] Iota.
(* (6.54u,0.14s) *)

(* 
Remark RRrefl : (x:A)(RR x x).
Proof refl_RR.

Remark RR01 : (x,y:A)(RR x y)->x=y\/(EX z:A |(R x z)/\(RR z y)).
Intros x y h. Case h.
Intro u. Left. Reflexivity.
Intros u v w h0 h1. Right. Exists v. Exact <?,?>{h0,h1}.
Save.

Remark RRtrans : (x,y,z:A)(RR x y)->(RR y z)->(RR x z).
Proof.
Intros x y z h.
Elim h; Clear h.
Trivial.
Intros u v w h h0 ih h1.
Exact (trans_RR u v z h (ih h1)).
Save.
*)

(* Theorem NL : (interpret f_NL).
Proof (NLplus RRrefl RRtrans RR01).
*)
*)
