Require Export axiom_types.

Section nnf_thms.

Variable M : MD.
Variable N : (etL M).
Variable R : (relT M N).
Local prp := (prop M N).
Local iE := (E M N R).
Variable em : EM .

Lemma raa : (p:(prp))(~(iE p)->False)->(iE p).
Proof.
Intros p H.
Elim (em M N R p).
Trivial.
Intro H0. Exact (False_ind (iE p)(H H0)).
Qed.

Lemma imp_to_or : (p,q:(prp))((iE p)->(iE q))->~(iE p)\/(iE q).
Proof.
Intros p q H.
Elim (em M N R p); Intro H0.
Right. Exact (H H0).
Left. Exact H0.
Qed.

Lemma DM1: (p,q:prp) ~((iE p)/\(iE q))->~(iE p)\/~(iE q).
Proof.
Intros p q H.
Apply (raa (f_Or (f_Not p)(f_Not q))).
Simpl. Intro H0.
Apply H0.
Left. Red. Intro H1.
Apply H0.
Right. Red. Intro H2.
Apply H.
Exact <(iE p),(iE q)>{H1,H2}.
Qed.

Lemma DM2: (p,q:Prop)~(p\/q)->~p/\~q.
Proof.
Intros p q H.
Split; Red; Intro H0; Apply H.
Left. Exact H0.
Right. Exact H0.
Qed.

Lemma NAEN: (k:nat;dp:(Dslct k M)->prp)
             ~((x:(Dslct k M))(iE (dp x)))
              ->(EX x:(Dslct k M)|~(iE (dp x))).
Proof.
Intros k dp H.
Apply (raa (f_Ex k [x:(Dslct k M)](f_Not (dp x)))). Intro H0.
Apply H. Intro x.
Apply (raa (dp x)). Intro H1.
Apply H0. Exists x. Exact H1.
Qed.

Local flip := [a:pol]if a then neg else pos.
Local piffp : (A:Prop)A<->A := [A:Prop]<(A->A),(A->A)>{([H:A]H),([H:A]H)}.

Lemma negation_flips :
 (p:prp;a:pol) ~(iE (nnf M N p a))<->(iE (nnf M N p (flip a))).
Proof.
Induction p. 
(* ATOM *)
Intros i b a. Case a.
(*pos*) Apply piffp.
(*neg*) Split.
Exact (raa (f_Atom i b)).
Exact [H:(R i b)][H':~(R i b)](H' H).
(* NOT *)
Intros p0 IH a.
Case a.
(*pos*) Exact (IH neg).
(*neg*) Exact (IH pos).
(* AND *)
Intros p1 IH1 p2 IH2 a.
Case a; Split; Intro H.
(*pos*)
(*->*) Elim (DM1 (nnf M N p1 pos)(nnf M N p2 pos) H); Intro H0.
Left. Exact (proj1 ? ? (IH1 pos) H0).
Right. Exact (proj1 ? ? (IH2 pos) H0).
(*<-*) Red. Intro H0. Elim H0. Intros H1 H2.
Elim H. Exact (proj2 ? ? (IH1 neg) H1). Exact (proj2 ? ? (IH2 neg) H2).
(*neg*)
(*->*) Split.
Apply (proj1 ? ? (IH1 neg)). Red. Intro H0. Apply H. Left. Exact H0.
Apply (proj1 ? ? (IH2 neg)). Red. Intro H0. Apply H. Right. Exact H0.
(*<-*) Red. Intro H0. Elim H. Intros H1 H2.
Elim H0. Exact (proj2 ? ? (IH1 neg) H1). Exact (proj2 ? ? (IH2 neg) H2).
(* OR *)
Intros p1 IH1 p2 IH2 a.
Case a; Split; Intro H.
(*pos*)
(*->*) Split.
Apply (proj1 ? ? (IH1 pos)). Red. Intro H0. Apply H. Left. Exact H0.
Apply (proj1 ? ? (IH2 pos)). Red. Intro H0. Apply H. Right. Exact H0.
(*<-*) Red. Intro H0. Elim H.  Intros H1 H2.
Elim H0. Exact (proj2 ? ? (IH1 pos) H1). Exact (proj2 ? ? (IH2 pos) H2).
(*neg*)
(*->*) Elim (DM1 (nnf M N p1 neg)(nnf M N p2 neg) H); Intro H0.
Left. Exact (proj1 ? ? (IH1 neg) H0).
Right. Exact (proj1 ? ? (IH2 neg) H0).
(*<-*) Red. Intro H0. Elim H0. Intros H1 H2.
Elim H. Exact (proj2 ? ? (IH1 pos) H1). Exact (proj2 ? ? (IH2 pos) H2).
(* IMPL *)
Intros p1 IH1 p2 IH2 a.
Case a; Split; Intro H.
(*pos*)
(*->*) Elim (DM2 (iE (nnf M N p1 neg))(iE (nnf M N p2 pos)) H).
Intros H1 H2.
Split. Exact (proj1 ? ? (IH1 neg) H1). Exact (proj1 ? ? (IH2 pos) H2).
(*<-*) Red. Intro H0. Elim H. Intros H1 H2.
Elim H0. Exact (proj2 ? ? (IH1 neg) H1). Exact (proj2 ? ? (IH2 pos) H2).
(*neg*)
(*->*) Elim (DM1 (nnf M N p1 pos)(nnf M N p2 neg) H); Intro H0.
Left. Exact (proj1 ? ? (IH1 pos) H0).
Right. Exact (proj1 ? ? (IH2 neg) H0).
(*<-*) Red. Intro H0. Elim H0. Intros H1 H2.
Elim H. Exact (proj2 ? ? (IH1 neg) H1). Exact (proj2 ? ? (IH2 pos) H2).
(* EX *)
Intros k dp IH a.
Case a; Split; Intro H.
(*pos*)
(*->*) Simpl. Intro x.
Apply (proj1 ? ? (IH x pos)).
Red. Intro H0.
Apply H. Exists x.
Exact H0.
(*<-*) Red. Intro H0. Elim H0. Intro x. Exact (proj2 ? ? (IH x pos) (H x)).
(*neg*)
(*->*) Elim (NAEN k [x:(Dslct k M)](nnf M N (dp x) neg) H).
Intros x H0. Exists x. Exact (proj1 ? ? (IH x neg) H0).
(*<-*) Red. Intro H0. Elim H. Intro x. Exact (proj2 ? ? (IH x pos) (H0 x)).
(* ALL *)
Intros k dp IH a.
Case a; Split; Intro H.
(*pos*)
(*->*) Elim (NAEN k [x:(Dslct k M)](nnf M N (dp x) pos) H).
Intros x H0. Exists x. Exact (proj1 ? ? (IH x pos) H0).
(*<-*) Red. Intro H0. Elim H. Intro x. Exact (proj2 ? ? (IH x neg) (H0 x)).
(*neg*)
(*->*) Simpl. Intro x.
Apply (proj1 ? ? (IH x neg)).
Red. Intro H0. Apply H. Exists x. Exact H0.
(*<-*) Red. Intro H0. Elim H0. Intro x. Exact (proj2 ? ? (IH x neg) (H x)).
Qed.

Theorem nnf_thm : (p:prp)
			(iE p)<->(iE (nnf M N p pos)).
(* due to cases (* NOT *) and (* IMPL *) we can't prove -> and <- separately *)
Proof. 
Induction p.
(* ATOM *)
Intros n a. Apply piffp.
(* NOT *)
Intros p0 IH. Split; Intro H.
(*->*) Apply (proj1 ? ? (negation_flips p0 pos)). Red.
Intro H0. Apply H. Exact (proj2 ? ? IH H0).
(*<-*) Simpl. Red. Intro H0.
Apply (proj2 ? ? (negation_flips p0 pos) H).
Exact (proj1 ? ? IH H0).
(* AND *)
Intros p1 IH1 p2 IH2. Split; Intro H; Elim H; Intros H1 H2; Split.
Exact (proj1 ? ? IH1 H1). Exact (proj1 ? ? IH2 H2).
Exact (proj2 ? ? IH1 H1). Exact (proj2 ? ? IH2 H2).
(* OR *)
Intros p1 IH1 p2 IH2. Split; Intro H; Elim H; Intro H0.
Left. Exact (proj1 ? ? IH1 H0).
Right. Exact (proj1 ? ? IH2 H0).
Left. Exact (proj2 ? ? IH1 H0).
Right. Exact (proj2 ? ? IH2 H0).
(* IMPL *)
Intros p1 IH1 p2 IH2. Split; Intro H.
(*->*) Elim (imp_to_or p1 p2 H); Intro H0.
Left. Apply (proj1 ? ? (negation_flips p1 pos)).
Red. Intro H1. Apply H0. Exact (proj2 ? ? IH1 H1).
Right. Exact (proj1 ? ? IH2 H0).
(*<-*) 
Simpl. Intro H0. Elim H; Intro H1.
Apply False_ind. Apply (proj2 ? ? (negation_flips p1 pos) H1).
Exact (proj1 ? ? IH1 H0).
Exact (proj2 ? ? IH2 H1).
(* EX *)
Intros k dp IH. Split; Intro H; Elim H; Intros x H0; Exists x.
Exact (proj1 ? ? (IH x) H0). Exact (proj2 ? ? (IH x) H0).
(* ALL *)
Intros k dp IH. Simpl. Split; Intros H x.
Exact (proj1 ? ? (IH x) (H x)). Exact (proj2 ? ? (IH x) (H x)).
Qed.

Theorem nnf_thm1 : (p:prp)(iE p)->(iE (nnf M N p pos)).
Proof.
Exact [p:prp](proj1 ? ? (nnf_thm p)).
Qed.

Theorem nnf_thm2 : (p:prp)(iE (nnf M N p pos))->(iE p).
Proof.
Exact [p:prp](proj2 ? ? (nnf_thm p)).
Qed.

End nnf_thms.
