Require Export axiom_types.

Section pnf_thms.

Variable M : MD.
Variable N : (etL M).
Variable R : (relT M N).
Local prp := (prop M N).
Local iE := (E M N R).
Variable ned:(NED M).

Lemma pnf_conj_r_lem1 : (p,q:prp)(iE p)/\(iE q)->(iE (pnf_conj_r M N p q)).
Proof.
Intros p q.
Elim q; Auto.
(* EX *)
Intros k dp IH H.
Elim H; Intros H1 H2.
Elim H2; Intros x H3.
Exists x.
Apply (IH x).
Split. Exact H1. Exact H3.
(* ALL *)
Simpl.
Intros k dp IH H x.
Elim H; Intros H1 H2.
Apply (IH x).
Split. Exact H1. Exact (H2 x).
Qed.

Lemma pnf_conj_lem1 : (p,q:prp)(iE p)/\(iE q)->(iE (pnf_conj M N p q)).
Proof.
Induction p.
Intros i a. Exact (pnf_conj_r_lem1 (f_Atom i a)).
Intros p0 IH. Exact (pnf_conj_r_lem1 (f_Not p0)).
Intros p1 IH1 p2 IH2. Exact (pnf_conj_r_lem1 (f_And p1 p2)).
Intros p1 IH1 p2 IH2. Exact (pnf_conj_r_lem1 (f_Or p1 p2)).
Intros p1 IH1 p2 IH2. Exact (pnf_conj_r_lem1 (f_Impl p1 p2)).
(* EX *)
Intros k dp IH q H.
Elim H. Intros H1 H2.
Elim H1. Intros x H0.
Exists x.
Apply (IH x).
Split. Exact H0. Exact H2.
(* ALL *)
Simpl. Intros k dp IH q H x.
Apply (IH x).
Elim H. Intros H1 H2.
Split. Exact (H1 x). Exact H2.
Qed.

Lemma pnf_disj_r_lem1 : (p,q:prp)(iE p)\/(iE q)->(iE (pnf_disj_r M N p q)).
Proof.
Induction q; Auto.
(* EX *)
Intros k dp IH H.
Elim H; Intro H0.
Exists (ned k).	Apply (IH (ned k)). (* NON-EMPTY DOMAINS REQUIRED! *)
Left. Exact H0.
Elim H0. Intros x H1.
Exists x. Apply (IH x).
Right. Exact H1.
(* ALL *)
Simpl. Intros k dp IH H x.
Apply (IH x).
Elim H; Intro H0.
Left. Exact H0.
Right. Exact (H0 x).
Qed.


Lemma pnf_disj_lem1 : (p,q:prp)(iE p)\/(iE q)->(iE (pnf_disj M N p q)).
Proof.
Induction p.
Intros i a. Exact (pnf_disj_r_lem1 (f_Atom i a)).
Intros p0 IH. Exact (pnf_disj_r_lem1 (f_Not p0)).
Intros p1 IH1 p2 IH2. Exact (pnf_disj_r_lem1 (f_And p1 p2)).
Intros p1 IH1 p2 IH2. Exact (pnf_disj_r_lem1 (f_Or p1 p2)).
Intros p1 IH1 p2 IH2. Exact (pnf_disj_r_lem1 (f_Impl p1 p2)).
(* EX *)
Intros k dp IH q H.
Elim H; Intro H0.
Elim H0; Intros x H1.
Exists x. Apply (IH x).
Left. Exact H1.
Exists (ned k). Apply (IH (ned k)).
Right. Exact H0.
(* ALL *)
Simpl. Intros k dp IH q H x.
Apply (IH x).
Elim H; Intro H0.
Left. Exact (H0 x).
Right. Exact H0.
Qed.

Theorem pnf_thm1 : (p:prp)(iE p)->(iE (pnf M N p)).
Proof.
Induction p; Auto.
(* AND *)
Intros p1 IH1 p2 IH2 H.
Apply (pnf_conj_lem1 (pnf M N p1)(pnf M N p2)).
Elim H; Intros H1 H2.
Split. Exact (IH1 H1). Exact (IH2 H2).
(* OR *)
Intros p1 IH1 p2 IH2 H.
Apply (pnf_disj_lem1 (pnf M N p1)(pnf M N p2)).
Elim H; Intro H0.
Left. Exact (IH1 H0).
Right. Exact (IH2 H0).
(* EX *)
Intros k dp IH H.
Elim H. Intros x H0.
Exists x. Exact (IH x H0).
(* ALL *)
Simpl. Intros k dp IH H x.
Exact (IH x (H x)).
Qed.

End pnf_thms.
