
open Pfedit
open Names
open Pp_control
open Pp
open Printer
open Vernacinterp
open Term
open Ast
open Tacmach
open Tactics
open Impargs
open Instantiate

let hd_app c = fst (destApplication c)

let args_app c = snd (destApplication c)

let constant s = Declare.global_absolute_reference (path_of_string s)

let coq_not_lazy = lazy (constant "Coq.Init.Logic.not")
let coq_or_lazy = lazy (constant "Coq.Init.Logic.or")
let coq_O_lazy = lazy (constant "Coq.Init.Datatypes.O")
let coq_S_lazy = lazy (constant "Coq.Init.Datatypes.S")
let coq_pair_lazy = lazy (constant "Coq.Init.Datatypes.pair")
let coq_eq_lazy = lazy (constant "Coq.Init.Logic.eq")

let rec print_tptp ft cnt t xxx =

  let coq_not = Lazy.force coq_not_lazy
  and coq_or = Lazy.force coq_or_lazy
  and coq_O = Lazy.force coq_O_lazy
  and coq_S = Lazy.force coq_S_lazy
  and coq_pair = Lazy.force coq_pair_lazy
  and coq_eq = Lazy.force coq_eq_lazy in

  let eq_skol t =
    match kind_of_term t with
      | IsConst (p,_) ->
          (string_of_id (basename p)) = xxx
      | IsMutInd ((p, 0), _) ->
          (string_of_id (basename p)) = xxx
      | IsVar i ->
          (string_of_id i) = xxx
      | _ -> false in

  let print_unknown t =
    [< 'sTR "\"" ; prterm t ; 'sTR "\"" >] in

  let print_id_upper i =
    [< 'sTR (String.uppercase (string_of_id i)) >] in

  let print_id_lower i =
    [< 'sTR (String.lowercase (string_of_id i)) >] in

  let rec eval_nat t =
    if eq_constr t coq_O then 0
    else if isApp t then
      let a = args_app t in
	if eq_constr (hd_app t) coq_S & Array.length a > 0 then
	  1 + eval_nat a.(0)
	else raise (Failure "eval_nat")
    else raise (Failure "eval_nat") in

  let print_nat t =
    [< 'iNT (eval_nat t) >] in
    
  let add_context n l =
    (match n with
	 Name i -> next_global_ident_away i l
       | Anonymous -> id_of_string "_") :: l in

  let filter_implicits t a =
    let rec filter l n b =
      match b with
	  c :: d ->
	    (match l with
		 m :: o ->
		   if m = n then filter o (n + 1) d
		   else c :: (filter l (n + 1) d)
	       | [] -> b)
	| [] -> [] in
      match kind_of_term t with
	  IsConst (p,_) -> filter (constant_implicits_list p) 1 a
	| _ -> a in

  let rec print_constr_clause v t =
    match kind_of_term t with
      | IsApp (f, a) ->
	    if eq_constr f coq_or & Array.length a = 2 then
	      [< print_constr_clause v a.(0) ; 'sTR "," ; 'fNL ;
		 print_constr_clause v a.(1) >]
	    else if eq_constr f coq_not & Array.length a = 1 then
	      [< 'sTR "--" ; print_constr v a.(0) >]
	    else
	      [< 'sTR "++" ; print_constr v t >]
      | IsProd (n, _, u) ->
	  print_constr_clause (add_context n v) u
      | IsLambda _ -> 
	  print_unknown t
      | _ -> [< 'sTR "+++" ; print_constr v t >]

  and print_constr_list v x = function
      [ ] -> [< >]
    | [ t ] -> [< print_constr v t >]
    | t :: (u :: a) ->
	[< print_constr v t ; x ; print_constr_list v x (u :: a) >]

  and print_constr v t =
    let t = collapse_appl t in
      match kind_of_term t with
	| IsApp (f, a) ->
	      if eq_constr f coq_S & Array.length a = 1 then
		try let n = eval_nat t in print_nat t
		with Failure "eval_nat" ->
		  [< 'sTR "s(" ; print_constr v a.(0) ; 'sTR ")" >]
	      else if eq_constr f coq_pair & Array.length a = 4 then
		hOV 0 [< print_constr v a.(2) ; 'sTR "," ; 'bRK(0,0) ;
			 print_constr v a.(3) >]
	      else if eq_constr f coq_eq & Array.length a = 3 then
		[< 'sTR "equal(" ;
		   hOV 0 [< print_constr v a.(1) ; 'sTR "," ; 'bRK(0,0) ;
			    print_constr v a.(2) >] ;  'sTR ")" >]
	      else if eq_skol f then
		let l = Array.to_list a in
		  [< print_constr v f;
		     hOV 0 (print_skol_list v [< 'sTR "," ; 'bRK(0,0) >]
			      4 (filter_implicits f l)) ;
		     'sTR ")" >]
	      else
		let l = Array.to_list a in
		  [< print_constr v f; 'sTR "(" ;
		     hOV 0 (print_constr_list v [< 'sTR "," ; 'bRK(0,0) >]
			      (filter_implicits f l)) ;
		     'sTR ")" >]
	| IsConst (p, _) ->
	    [< print_id_lower (basename p) >]
	| IsMutInd ((p, n), _) ->
	    if n = 0 then
	      [< print_id_lower (basename p) >]
	    else
	      prterm t
	| IsConst _ | IsMutInd _ | IsMutConstruct _ ->
	    if eq_constr t coq_O then
	      print_nat t
	    else prterm t
	| IsProd (n, _, u) -> print_constr (add_context n v) u
	| IsVar i -> print_id_lower i
	| IsRel n -> [< print_id_upper (List.nth v (n - 1)) >]
	| _ -> print_unknown t
    
  and print_skol_list v x n a =
    if n = 0 then
      [< 'sTR "("; print_constr_list v x a >]
    else
      [< 'sTR "_"; print_skol v (List.hd a);
	 print_skol_list v x (n-1) (List.tl a) >]

  and print_skol v t =
    match kind_of_term t with
      | IsApp (f, a) ->
	    if eq_constr f coq_S then
	      try let n = eval_nat t in print_nat t
	      with Failure "eval_nat" -> print_unknown t
	    else if eq_constr f coq_pair & Array.length a = 4 then
	      hOV 0 [< print_skol v a.(2) ; 'sTR "_" ;
		       print_skol v a.(3) >]
	    else print_unknown t
      | IsMutConstruct _ when eq_constr t coq_O ->
	  print_nat t
      | _ -> print_unknown t in
	    
    mSG_with ft [< hOV 4 [< 'sTR "input_clause(coq_clause" ;
			    (* 'sTR (if cnt < -1 then "_clause"
				  else "_hyp") ; *)
			    (if cnt >= 0 then [< 'sTR "_" ; 'iNT cnt >]
			     else [< >]) ;
			    'sTR ",axiom," ; 'fNL ;
			    hOV 1 [< 'sTR "[" ; print_constr_clause [] t >] ;
			    'bRK(0,0) ; 'sTR "])." >] ; 'fNL; 'fNL >]
      
let tptp ft a xxx =
  let env = Global.env() in 
  let t = Astterm.interp_constr Evd.empty env a in
    match kind_of_term t with
	IsConst c -> print_tptp ft (-2) (constant_value env c) xxx
      | IsVar _ -> print_tptp ft (-2) (Typing.type_of env Evd.empty t) xxx
      | _ -> print_tptp ft (-2) t xxx

let rec tptp_hypos ft n a xxx =
  let b = strip_outer_cast a in
    match kind_of_term b with
	IsProd(_, d, r) -> print_tptp ft n d xxx; tptp_hypos ft (n+1) r xxx;
      | _ -> ()

let tptp_goal ft xxx =
  let a = pf_concl (nth_goal_of_pftreestate 1 (get_pftreestate ())) in
    tptp_hypos ft (-1) a xxx

let file_fd x n =
  Unix.openfile n
    (if x = 0 then [ Unix.O_WRONLY; Unix.O_TRUNC; Unix.O_CREAT ]
     else [ Unix.O_WRONLY; Unix.O_APPEND; Unix.O_CREAT ])
    0o666

let file_ft fd =
  let ch = Unix.out_channel_of_descr fd in
    Format.formatter_of_out_channel ch

let _ =
  vinterp_add "tptp"
     (function
	 [VARG_CONSTR a] ->
	   (fun () -> tptp std_ft a "")
       | [VARG_NUMBER x; VARG_STRING n; VARG_CONSTR a] ->
	   (fun () -> let fd = file_fd x n in let ft = file_ft fd in
	      tptp ft a ""; Unix.close fd)
       | [] ->
	   (fun () -> tptp_goal std_ft "")
       | [VARG_NUMBER x; VARG_STRING n] ->
	   (fun () -> let fd = file_fd x n in let ft = file_ft fd in
	      tptp_goal ft ""; Unix.close fd)
       | [VARG_CONSTR a; VARG_STRING xxx] ->
	   (fun () -> tptp std_ft a xxx)
       | [VARG_NUMBER x; VARG_STRING n; VARG_CONSTR a; VARG_STRING xxx] ->
	   (fun () -> let fd = file_fd x n in let ft = file_ft fd in
	      tptp ft a xxx; Unix.close fd)
       | [VARG_STRING xxx] ->
	   (fun () -> tptp_goal std_ft xxx)
       | [VARG_NUMBER x; VARG_STRING n; VARG_STRING xxx] ->
	   (fun () -> let fd = file_fd x n in let ft = file_ft fd in
	      tptp_goal ft xxx; Unix.close fd)
       | _ -> Util.anomaly "TPTP")

